# -------------------------------------------------------------------------------
# - Copyright (c) 2021 Arista Networks, Inc. All rights reserved.
# -------------------------------------------------------------------------------
# - Author:
# -   fdk-support@arista.com
# -
# - Description:
# -   Factories to create DUT fixtures
# -
# -   Licensed under BSD 3-clause license:
# -     https://opensource.org/licenses/BSD-3-Clause
# -
# - Tags:
# -   license-bsd-3-clause
# -
# -------------------------------------------------------------------------------
import logging
import re
import pytest
from .wrappers import CLI, xapi


def create_dut_fixture(name):
    @pytest.fixture(name=f"{name}")
    def _dut(request):
        skipper = request.getfixturevalue(f"{name}_skipper")
        skipper(request.node)

        class Dut:
            def __getattr__(self, attr):
                return request.getfixturevalue(f"{name}_{attr}")

        yield Dut()

    return _dut


def create_eapi_fixture(name):
    @pytest.fixture(scope="session", name=f"{name}_eapi")
    def _eapi(eapi_enabled, request):
        hostname = request.getfixturevalue(f"{name}_hostname")
        ssh = request.getfixturevalue(f"{name}_ssh")
        with eapi_enabled(hostname, ssh, "http") as eapi:
            yield eapi

    return _eapi


def create_hostname_fixture(name):
    @pytest.fixture(scope="session", name=f"{name}_hostname")
    def _hostname(request):
        yield request.getfixturevalue(f"{name}_info")["hostname"]

    return _hostname


def create_skipper_fixture(name):
    @pytest.fixture(scope="session", name=f"{name}_skipper")
    def _skipper(request):
        def skipper(node):
            allowed_os = set()

            # Restrict SKUs
            for marker in node.iter_markers():
                if marker.name in {"mos", "eos"}:
                    allowed_os.add(marker.name)

                elif marker.name == "only_device_type":
                    pattern = marker.args[0]
                    sku = request.getfixturevalue(f"{name}_sku")
                    if not re.search(pattern, sku):
                        pytest.skip(
                            f"Skipped on this SKU: {sku} (only runs on {pattern})"
                        )

                elif marker.name == "skip_device_type":
                    pattern = marker.args[0]
                    sku = request.getfixturevalue(f"{name}_sku")
                    if re.search(pattern, sku):
                        pytest.skip(f"Skipped on this SKU: {sku}")

            if allowed_os:
                dut_ssh = request.getfixturevalue(f"{name}_ssh")
                if dut_ssh.cli_flavor not in allowed_os:
                    pytest.skip(f"cannot run on platform {dut_ssh.cli_flavor}")

        return skipper

    return _skipper


def create_sku_fixture(name):
    @pytest.fixture(scope="session", name=f"{name}_sku")
    def _sku(request):
        ssh = request.getfixturevalue(f"{name}_ssh")
        assert ssh.cli_flavor in {"eos", "mos"}
        output = ssh.sendcmd("show version", timeout=300)
        matcher = re.search(r"(DCS-7.*)", output)
        logging.info("Got SKU: %s", matcher.group(1))
        yield matcher.group(1)

    return _sku


def create_softened_fixture(name):
    @pytest.fixture(scope="session", name=f"{name}_softened")
    def _softened(dut_ssh):
        # Set up a standard operating environment.
        dut_ssh.sendcmds(
            """
            enable
                configure
                    username admin privilege 15 nopassword
                    aaa authorization exec default local
        """
        )
        yield

        # FIXME: Return the running config to its pre-test state.

    return _softened


def create_ssh_fixture(name):
    @pytest.fixture(scope="session", name=f"{name}_ssh")
    def _ssh(request):
        try:
            ssh = CLI(f"ssh://{request.getfixturevalue(f'{name}_hostname')}")
            ssh.login()
        except Exception as exc:
            logging.error("Failed to create ssh fixture and log in")
            raise exc
        if ssh.cli_flavor == "mos":
            ssh.sendcmd("enable")
        yield ssh

    return _ssh


def create_xapi_fixture(name):
    @pytest.fixture(scope="session", name=f"{name}_x")
    def _xapi(request):
        yield xapi(request.getfixturevalue(f"{name}_eapi"))

    return _xapi



# -------------------------------------------------------------------------------
# - Copyright (c) 2021-2022 Arista Networks, Inc. All rights reserved.
# -------------------------------------------------------------------------------
# - Author:
# -   fdk-support@arista.com
# -
# - Description:
# -   Tests for the xapi fixture.
# -
# -   Licensed under BSD 3-clause license:
# -     https://opensource.org/licenses/BSD-3-Clause
# -
# - Tags:
# -   license-bsd-3-clause
# -
# -------------------------------------------------------------------------------

import pytest
import pytest_netdut

pytest_plugins = ["pytest_mock"]


@pytest.fixture
def mock_eapi(mocker):
    def sendcmds(cmds):
        if cmds == ["show version"]:
            return [
                {
                    "architecture": "x86_64",
                    "bootupTimestamp": 1628827627.0,
                    "configMacAddress": "00:00:00:00:00:00",
                    "hardwareRevision": "01.03",
                    "hwMacAddress": "7c:53:4a:d3:67:80",
                    "imageFormatVersion": "01.00",
                    "internalBuildId": "43757e25-1545-4d6c-b200-c768cb3053fb",
                    "internalVersion": "4.26.1FX-7130-22950577.4261FX7130",
                    "isIntlVersion": False,
                    "memFree": 6021356,
                    "memTotal": 8083948,
                    "mfgName": "Metamako",
                    "modelName": "DCS-7130-48L-R",
                    "serialNumber": "M48L-A3-54119-2",
                    "systemMacAddress": "7c:53:4a:d3:67:80",
                    "uptime": 29857.29,
                    "version": "4.26.1FX-7130",
                },
            ]

    eapi = mocker.MagicMock()
    eapi.sendcmds = mocker.MagicMock(side_effect=sendcmds)
    yield eapi


@pytest.fixture
def mock_xapi(mock_eapi):
    yield pytest_netdut.xapi(mock_eapi)


def test_xapi_show_version(mock_xapi):
    result = mock_xapi.show_version()
    assert result["version"] == "4.26.1FX-7130"



# -------------------------------------------------------------------------------
# - Copyright (c) 2021-2022 Arista Networks, Inc. All rights reserved.
# -------------------------------------------------------------------------------
# - Author:
# -   fdk-support@arista.com
# -
# - Description:
# -   Tests for the wait_for function.
# -
# -   Licensed under BSD 3-clause license:
# -     https://opensource.org/licenses/BSD-3-Clause
# -
# - Tags:
# -   license-bsd-3-clause
# -
# -------------------------------------------------------------------------------

import time
import pexpect
import pytest

pytest_plugins = ["pytest_netdut"]


def test_wait_for(wait_for):
    """Make sure that pytest reports a timeout."""
    # The timeout for this particular test
    TIMEOUT = 1
    # The hard-coded TIMEOUT_LEEWAY_FACTOR constant must be the same as the
    # one similarly defined in the wait_for function (in utils.py).
    TIMEOUT_LEEWAY_FACTOR = 2

    def good_fn():
        time.sleep(0.5 * TIMEOUT)
        return True

    def slow_fn():
        time.sleep((1 + TIMEOUT_LEEWAY_FACTOR) * TIMEOUT / 2)
        return True

    def bad_fn():
        time.sleep(1.1 * TIMEOUT_LEEWAY_FACTOR * TIMEOUT)
        return False

    # Test a good function which completes within "timeout" (i.e. does not time out)
    # T < timeout
    wait_for(good_fn, timeout=TIMEOUT)

    # Test a slow function which completes after "timeout" but before "TIMEOUT_LEEWAY_FACTOR * timeout"
    # timeout <= T < TIMEOUT_LEEWAY_FACTOR * timeout
    with pytest.raises(pexpect.TIMEOUT, match=r"succeeded"):
        wait_for(slow_fn, timeout=TIMEOUT)

    # Test a bad function which doesn't complete within "TIMEOUT_LEEWAY_FACTOR * timeout"
    # T >= TIMEOUT_LEEWAY_FACTOR * timeout
    with pytest.raises(pexpect.TIMEOUT, match=r"unsuccessful"):
        wait_for(bad_fn, timeout=TIMEOUT)



import logging
import pprint
import pytest


@pytest.fixture
def sleeper_daemon(dut, wait_for):
    # Start a dummy daemon that sleeps
    dut.eapi.sendcmds(
        """
        enable
            configure
                daemon sleeper
                    exec /usr/bin/sleep 10
                    no shutdown
    """
    )

    # Wait for the OS to register the config and add the daemon.
    wait_for(
        lambda: "sleeper" in dut.eapi.sendcmds(["enable", "show daemon"])[-1]["daemons"]
    )

    # Yield the name of the daemon for the test to use.
    yield "sleeper"

    dut.eapi.sendcmds(
        """
        enable
            configure
                no daemon sleeper
    """
    )

    # Wait for the OS to register the config and remove the daemon.
    wait_for(
        lambda: "sleeper"
        not in dut.eapi.sendcmds(["enable", "show daemon"])[-1]["daemons"]
    )


def test_sendcmds(dut, sleeper_daemon, wait_for):

    # This local functon determines whether the daemon has started.
    # We call it below using wait_for.
    def daemon_has_started():
        daemon_info = dut.eapi.sendcmds(["enable", f"show daemon {sleeper_daemon}"])[-1]
        return daemon_info["daemons"][sleeper_daemon]["starttime"] != 0.0

    # Wait for the Daemon to start. If it doesn't within 10s, the test fails.
    wait_for(daemon_has_started, timeout=10.0)

    # Snapshot the info about the daemon
    daemon_info = dut.eapi.sendcmds(["enable", f"show daemon {sleeper_daemon}"])[-1][
        "daemons"
    ][sleeper_daemon]

    # Check that our uptime is still non-zero
    assert daemon_info["starttime"] != 0.0
    assert daemon_info["uptime"] != 0.0
    assert daemon_info["running"]

    # Print the PID, so we can show we really created a daemon.
    logging.info(f"Daemon was started and the PID is {daemon_info['pid']}")



import logging
import pytest
import time


@pytest.fixture
def my_test_harness(dut):
    dut.ssh.sendcmds(
        ["enable", "configure", "banner motd\nMy test is running, hands off!\nEOF\n"]
    )
    yield
    dut.ssh.sendcmds(["enable", "configure", "no banner motd"])


def test_mytest(my_test_harness):
    logging.info("Sleeping to simulate running a test here")
    time.sleep(5)



# -------------------------------------------------------------------------------
# - Copyright (c) 2021-2022 Arista Networks, Inc. All rights reserved.
# -------------------------------------------------------------------------------
# - Author:
# -   fdk-support@arista.com
# -
# - Description:
# -   Utility functions to use in pytest-netdut.
# -
# -   Licensed under BSD 3-clause license:
# -     https://opensource.org/licenses/BSD-3-Clause
# -
# - Tags:
# -   license-bsd-3-clause
# -
# -------------------------------------------------------------------------------

import contextlib
import time
import pexpect
import pytest


@pytest.fixture(scope="session")
def wait_for():
    """A fixture that returns a function that can be called to wait for something to happen.

    This is useful waiting for configuration to be applied asynchronously.

    Examples:
        >>> wait_for(lambda: dut.eapi.sendcmd(f"show {appname} status")["running"])
    """

    def _wait_for(func, timeout=30, period=0.1, suppress=None):
        """Wait up to 'timeout' seconds for the successful completion of 'func'."""
        # Let the elapsed time within this function be T seconds.
        # There are 3 scenarios:

        # 1.  T < 'timeout'
        #    'func' returns successfully.
        #    This function returns.

        # 2. 'timeout' <= T < TIMEOUT_LEEWAY_FACTOR * 'timeout'
        #    'func' returns successfully.
        #    This function raises an exception with an error message.

        # 3. T >= TIMEOUT_LEEWAY_FACTOR * 'timeout'
        #    'func' returns unsuccessfully.
        #    This function raises an exception with an error message.

        # The hard-coded timeout_leeway_factor constant must be the same as the
        # one similarly defined in the test_wait_for function (in test_wait_for.py).
        # This constant is used to allow for some leeway in the reporting of a
        # 'func' which was successful, but whose duration exceeded the requested
        # 'timeout', i.e. 'func' returned successfully after a time T where
        # 'timeout' <= T < timeout_leeway_factor * 'timeout'.
        timeout_leeway_factor = 2

        # Alert the user if the polling period is greater than the requested timeout.
        assert timeout > period

        if suppress is None:
            suppress = []
        start = time.time()
        while True:
            with contextlib.suppress(*suppress):
                result = func()
                if result:
                    elapsed = time.time() - start
                    if elapsed > timeout:
                        raise pexpect.TIMEOUT(
                            f"Timed out waiting for {func} after {timeout} seconds, "
                            f"but eventually succeeded after {elapsed} seconds."
                        )
                    return result
            if time.time() - start > timeout_leeway_factor * timeout:
                raise pexpect.TIMEOUT(
                    f"Function {func} was unsuccessful after the requested {timeout} "
                    f"second timeout, and was also unsuccessful after being allowed "
                    f"a leeway of {timeout_leeway_factor * timeout} seconds."
                )
            time.sleep(period)

    yield _wait_for



# -------------------------------------------------------------------------------
# - Copyright (c) 2021-2022 Arista Networks, Inc. All rights reserved.
# -------------------------------------------------------------------------------
# - Author:
# -   fdk-support@arista.com
# -
# - Description:
# -   Tests for the netdut fixtures.
# -
# -   Licensed under BSD 3-clause license:
# -     https://opensource.org/licenses/BSD-3-Clause
# -
# - Tags:
# -   license-bsd-3-clause
# -
# -------------------------------------------------------------------------------


def test_dut_fixtures(testdir):
    """Make sure that pytest accepts our fixture."""

    # create a temporary conftest module
    testdir.makeconftest(
        """
        pytest_plugins = ["pytest_netdut"]
    """
    )

    # create a temporary pytest test module
    testdir.makepyfile(
        """
        def test_sth(dut):
            assert dut.info["hostname"] == "baz01"
            assert dut.info["console_url"] == "telnet://bernie:1001"
    """
    )

    # run pytest with the following cmd args
    result = testdir.runpytest("--device=baz01", "--console=telnet://bernie:1001", "-v")

    # fnmatch_lines does an assertion internally
    result.stdout.fnmatch_lines(
        [
            "*::test_sth PASSED*",
        ]
    )

    # make sure that we get a '0' exit code for the testsuite
    assert result.ret == 0


def test_help_message(testdir):
    testdir.makeconftest(
        """
        pytest_plugins = ["pytest_netdut"]
    """
    )

    result = testdir.runpytest(
        "--help",
    )
    # fnmatch_lines does an assertion internally
    result.stdout.fnmatch_lines(
        [
            "netdut:",
            "*--device=DUT_HOSTNAME",
            "*DUT hostname",
            "*--console=DUT_CONSOLE_URL",
            "*DUT console URL",
        ]
    )



# -------------------------------------------------------------------------------
# - Copyright (c) 2021 Arista Networks, Inc. All rights reserved.
# -------------------------------------------------------------------------------
# - Author:
# -   fdk-support@arista.com
# -
# - Description:
# -   Convenience wrappers around pexpect for CLIs.
# -
# -   Licensed under BSD 3-clause license:
# -     https://opensource.org/licenses/BSD-3-Clause
# -
# - Tags:
# -   license-bsd-3-clause
# -
# -------------------------------------------------------------------------------

# pylint: disable=consider-using-f-string

from __future__ import absolute_import, print_function

import os
import re
import subprocess
import six.moves.urllib.parse
import sys
import time

import pexpect
import six
from pexpect import TIMEOUT, ExceptionPexpect
from six.moves import range


class spawn(pexpect.spawn):
    def __str__(self):
        spawned = pexpect.spawn.__str__(self).splitlines()
        spawned.append("_prompt: " + repr(getattr(self, "_prompt", None)))
        return "\n".join(spawned)


class Shell(spawn):
    _control_code_re = r"(?:\x1B(?:.|[@-_][0-?]*[ -/]*[@-~]))*"  # Match control codes

    # Match device hostname or bash prompt (requires at least one alphabet)
    _hostname_re = r"([a-zA-Z]+[\w\-\.]*)"
    _mode_and_path_re = (
        r"("  # Match
        r">"  # cli non-priv mode
        r"|"  # or
        # cli priv and config modes (only simple config mode names are allowed)
        r"((\([\w\-\.\,\/]+\))?#)"
        r"|"  # or
        r"(:[\w\-\.\/~]+(#|\$))"  # bash working dir and shell type (bash, sudo bash)
        # End optional (bash-4.4# is a special case that matches cli priv prompt)
        r")"
    )
    _prompt_regex = re.compile(_control_code_re + _hostname_re + _mode_and_path_re)

    def syncprompt(self):
        self.expect(self._prompt_regex)
        # Use captured serial number / hostname for finding the prompt
        self._prompt = re.compile(
            self._control_code_re + self.match.group(1) + self._mode_and_path_re
        )

    def prompt(self, timeout=-1, reset=0):
        if reset:
            self.syncprompt()
        else:
            self.expect(self._prompt, timeout)
        return self.before.replace("\r\n", "\n")

    def sendcmd(self, cmd="", timeout=-1, reset=0, wait=True):
        if not isinstance(self.after, six.string_types) or not self._prompt.match(
            self.after
        ):
            raise Exception(
                "Cannot find prompt. Refusing to send command.",
                self.after,
                self._prompt.pattern,
            )

        self.sendcmd_unchecked(cmd)

        if wait:
            return self.prompt(timeout, reset)

        return None

    def sendcmd_unchecked(self, cmd=""):
        # "en" is a common abbreviation for enable.
        # We special-case this particular command.
        if cmd == "en":
            cmd = "enable"

        # self.sendline() appends os.linesep, so strip any line endings from
        # the end of the command to prevent the prompt from getting out of
        # sync.
        cmd = cmd.rstrip(os.linesep)

        self.sendline(cmd)

        try:
            self.expect(re.escape(cmd), timeout=10)
        except TIMEOUT:
            # This deals with long commands on the serial console
            if not cmd.startswith(
                self.before[  # pylint: disable=unsubscriptable-object
                    : self.before.index("\r")
                ].strip()
            ):
                raise

        # This deals with disappearing newlines on tty
        try:
            self.expect("\r\n", timeout=10)
        except TIMEOUT:
            self.sendline()
            self.expect("\r\n", timeout=10)

    # This deals with disappearing SIGINTs on tty
    def sendintr(self, retries=5, timeout=5, reset=0, wait=True):
        for i in range(retries):
            try:
                pexpect.spawn.sendintr(self)
                self.expect(re.escape("^C"), timeout)
                if wait:
                    return self.prompt(timeout, reset)
                break
            except TIMEOUT:
                if i == (retries - 1):
                    raise
        return None

    def is_at_prompt(self):
        if not self.after or not self._prompt:
            return False
        if not isinstance(self.after, six.string_types):
            return False
        match = re.match(self._prompt, self.after)
        return match is not None


class CLI(Shell):
    def __init__(  # pylint: disable=dangerous-default-value,too-many-arguments
        self,
        url,
        username="admin",
        password="",
        timeout=30,
        enable_cli_timeout=False,
        cli_flavor="mos",
        extra_args=[],
    ):
        o = six.moves.urllib.parse.urlparse(url)  # pylint: disable=invalid-name
        self.cmd = o.scheme
        if self.cmd == "ssh":
            self.args = ["%s@%s" % (username, o.hostname)]
            self.args += ["-o LogLevel ERROR"]
            self.args += ["-o StrictHostKeyChecking no"]
            self.args += ["-o UserKnownHostsFile /dev/null"]
            # default is whatever TCP timeout at OS level
            # self.args += ['-o ConnectTimeout 10']
            # self.args += ['-o ServerAliveCountMax 3']   # default 3
            # self.args += ['-o ServerAliveInterval 30']  # default 0
            self.args += ["-p %s" % (o.port or "22")]
            self.args += extra_args
        elif self.cmd in ("tcp", "telnet"):
            self.cmd = "mc"
            self.args = ["%s:%s" % (o.hostname, o.port)]
            self.args += extra_args
        elif self.cmd == "conserver":
            # Sample console spec: conserver://<netloc>/<path>
            # Executes:  console -f -M <netloc> -e^^^^ <path>
            conserver = o.netloc
            conport = o.path.strip("/")
            self.cmd = "console"
            self.args = ["-f", "-M", conserver, "-e^^^^", conport]
            self.args += extra_args
        else:
            self.args = [o.hostname, str(o.port)]
            self.args += extra_args
        Shell.__init__(
            self,
            self.cmd,
            self.args,
            timeout,
            dimensions=(60, 500),
            encoding="utf-8" if six.text_type == str else None,
        )
        self.logfile_read = sys.stdout
        self.delayafterterminate = 1
        self.delayafterclose = 1
        self.username = username
        self.password = password
        self.enable_cli_timeout = enable_cli_timeout
        self.cli_timeout = 0
        self.cli_flavor = cli_flavor
        self.device_generation = None
        self.plm_wd_supported = None
        self.serial = None
        self.micro_version = None

    def login(self, timeout=30):  # noqa: MC0001
        if self.cmd != "ssh":
            time0 = time.time()
            i = 0
            while True:
                if time.time() - time0 > timeout:
                    raise TIMEOUT("failed to login")

                index = self.expect(
                    [
                        TIMEOUT,
                        self.login_prompt_re_mos,
                        self.login_prompt_re_eos,
                        self.login_prompt_re_aboot,
                    ],
                    timeout=2,
                )

                print(" -- Got a login prompt index of {} -- ".format(index))

                if index == 1:
                    self.cli_flavor = "mos"
                    break

                if index == 2:
                    self.cli_flavor = "eos"
                    break

                if index == 3:
                    self.cli_flavor = "aboot"
                    break

                # Note that blindly sending EOFs and INTRs may break
                # some of the init scripts if the device is rebooting.
                if i % 2 == 0:
                    self.sendeof()
                else:
                    pexpect.spawn.sendintr(self)

                i += 1

            if self.cli_flavor != "aboot":
                self.sendline(self.username)
                self.expect(self.username + "(\r)?\r\n", timeout=10)
            else:
                self.sendline()

        print("\n -- Logged in. Sending password, etc. -- \n")

        if self.cli_flavor != "aboot":
            # This should also handle the case where a password prompt is presented
            # because a TACACS server has been configured even though there is no
            # password actually configured for the user.
            index = self.expect(
                [TIMEOUT, "Last login:.*\r\n", "Password:", "Aboot#"], timeout=10
            )
            if index == 2:
                self.sendline(self.password)

        print("\n -- Waiting for prompt -- \n")

        try:
            self.prompt(reset=1)
        except TIMEOUT:
            # Check if the CLI is busted or if previous tests left a password set
            try:
                index = self.expect(
                    ["Traceback", "Login incorrect", "Password:"], timeout=10
                )
                if index == 0:
                    raise Exception(  # pylint: disable=raise-missing-from
                        "CLI is busted, try logging in as root!"
                    )
                if index == 1:
                    self.expect("login:", timeout=2)
                    self.sendline(self.username)
                    self.expect(self.username + "(\r)?\r\n", timeout=10)
                    self.expect("Password:", timeout=10)
                    self.sendline("opensesame")  # This is used in some tests.
                    self.expect("Last login:.*\r\n", timeout=10)
                elif index == 2:
                    self.sendline("opensesame")  # This is used in some tests.
                    self.expect("Last login:.*\r\n", timeout=10)
                self.expect("Traceback", timeout=10)
                raise Exception(  # pylint: disable=raise-missing-from
                    "CLI is busted, try logging in as root!"
                )
            except TIMEOUT:
                self.prompt(reset=1)

        # Sometime there are weird control characters printed as "^[[0n"
        # on the first line that escape the control code filtering.
        # Resync just in case.
        self.sendline()
        self.prompt(reset=1)

        # Do standard setup for MOS/EOS.
        if self.cli_flavor != "aboot" and self.username != "root":
            # Sendcmd via Shell directly to avoid messing with cli_timeouts
            # since we do not yet know for sure what OS the device is running.
            show_version = self.sendcmd_simple("show version", timeout=10)
            matcher = re.search("Serial number:[ \t]*(.*)", show_version)
            if matcher:
                self.serial = matcher.group(1)

            matcher = re.search(
                r"System management controller version: (\d+)", show_version
            )
            if matcher:
                self.micro_version = matcher.group(1)

            # More reliably determine the CLI flavor
            # MOS does not have a 'Hardware version' field in "show version'
            matcher = re.search("Hardware version:", show_version)
            if matcher:
                self.cli_flavor = "eos"
            else:
                self.cli_flavor = "mos"

            # Check if previous tests left a password set
            try:
                self.sendcmd_simple("enable")
            except TIMEOUT:
                self.expect("Password:", timeout=10)
                self.sendline("opensesame")  # This is used in some tests.
                output = self.prompt(timeout=timeout, reset=1)
                self.process_output(output)

            self.sendcmd_simple(
                "bash echo ===> px Determined the {} CLI flavor".format(self.cli_flavor)
            )

            if self.cli_flavor == "mos":
                self.sendcmd_simple("set debug 1", timeout=10)

                # Set default cli timeout which
                # is the 2x the command timeout
                if self.enable_cli_timeout:
                    self.set_cli_timeout()

                # Determine device generation
                try:
                    output = self.sendcmd(
                        "bash python -m hal property chassis chassis_gen"
                    )
                    self.device_generation = int(output.strip())
                except Exception:  # pylint: disable=broad-except
                    plm_ver = self.sendcmd("bash i2cget -f -y 1 0x77 0x8 w").strip()
                    self.device_generation = 1 if int(plm_ver, 16) < 0x200 else 2

                # If Gen2 device, determine PLM watchdog support by
                # querying register PLM_VER_PATCH(0x7e)
                try:
                    if self.device_generation == 2:
                        plm_ver_patch = self.sendcmd(
                            "bash i2cget -f -y 1 0x77 0x7e b"
                        ).strip()
                        self.plm_wd_supported = plm_ver_patch != "0xdb"
                except Exception:  # pylint: disable=broad-except
                    pass

            self.sendcmd("show clock", timeout=timeout)

        if self.cmd != "ssh":
            # EOS and Aboot wrap lines at 80 chars on the console by default
            # and this breaks the command line echo regex, so tell it we have
            # a really wide terminal.  If our command lines ever hit 1000
            # chars we're doing something wrong :)
            if self.cli_flavor == "aboot":
                self.sendcmd("stty cols 1000")
            elif self.cli_flavor == "eos":
                self.sendcmd("bash stty cols 1000")

    def _can_insert_cli_timeout(self):
        return self.is_at_prompt()

    def set_cli_timeout(self, cli_timeout=-1):
        """Set the CLI timeout with the set timeout command.

        Does nothing if the timeout was already set.
        :param cli_timeout: CLI timeout in seconds. 0 is no timeout, -1 is the
        default (2 * self.timeout).
        :return:
        """
        if self.cli_flavor != "mos" or not self.enable_cli_timeout:
            return

        # If a CLI timeout is not specified set the default cli
        # timeout which is the 2x the command timeout.
        if cli_timeout == -1:
            cli_timeout = 2 * self.timeout

        if cli_timeout != self.cli_timeout:
            # We are at the prompt, safe to insert a timeout command
            if self._can_insert_cli_timeout():
                # Cache the cli timeout
                self.cli_timeout = cli_timeout

                # Failsafe in case the provided timeout is funky
                if cli_timeout > 0:
                    self.sendcmd_simple("set timeout {}".format(cli_timeout))
                else:
                    self.sendcmd_simple("no timeout")
            else:
                raise Exception("Not at CLI prompt")

    def sendcmd_simple(self, cmd="", timeout=-1, reset=0, wait=True):
        if wait:
            output = Shell.sendcmd(self, cmd, timeout, reset, wait)
            return self.process_output(output)

        Shell.sendcmd(self, cmd, timeout, reset, wait)
        return None

    def sendcmd(self, cmd="", timeout=-1, reset=0, wait=True):
        if timeout == -1:
            timeout = self.timeout

        cli_timeout = 0
        shell_timeout = 2 * timeout

        if self.enable_cli_timeout:
            # Timeout does not work as expected in sub-shells, stop it here.
            if wait and not cmd.strip().startswith(("bash", "python")):
                cli_timeout = 2 * timeout
                shell_timeout = 3 * timeout

            self.set_cli_timeout(cli_timeout)

        if wait:
            start_time = time.time()
            output = self.sendcmd_simple(cmd, shell_timeout, reset, wait)
            elapsed_time = time.time() - start_time

            if elapsed_time > timeout:
                if self.is_at_prompt():
                    raise CommandTimeoutError(cmd, timeout, elapsed_time)

                raise PromptTimeoutError(cmd, timeout, elapsed_time)

            return output

        self.sendcmd_simple(cmd, shell_timeout, reset, wait)
        return None

    @property
    def login_prompt_re(self):
        if self.cli_flavor == "mos":
            return self.login_prompt_re_mos
        if self.cli_flavor == "aboot":
            return self.login_prompt_re_aboot

        assert self.cli_flavor == "eos"
        return self.login_prompt_re_eos

    @property
    def login_prompt_re_eos(self):
        # "\S+" here is the host name, which is unset as EOS does not
        # Configure the management network by default
        return r"\r\n\S+ login: "

    @property
    def login_prompt_re_mos(self):
        return r"Metamako MOS \S+ \S+ (/dev/)?ttyS0\r\n\s*\r\n.*login: "

    @property
    def login_prompt_re_aboot(self):
        return r"Aboot# "

    def strip_control_codes(self, value):
        value = re.sub(r"\x1B([@-_][0-?]*[ -/]*[@-~]|.)", "", value)
        value = value.replace("\r", "")
        return value

    def process_output(self, output):
        output = self.strip_control_codes(output)
        matcher = re.search("^% (.*)$", output, re.M)
        if matcher:
            error = ExceptionPexpect(matcher.group(1))
            error.output = output
            raise error
        return output


class FlakyTimeoutError(Exception):
    def __init__(self, cmd, timeout, actual_time):
        super(FlakyTimeoutError, self).__init__(  # pylint: disable=super-with-arguments
            cmd, timeout, actual_time
        )
        self.cmd = cmd
        self.timeout = timeout
        self.actual_time = actual_time

    def __repr__(self):
        return "{}(cmd={!r}, timeout={!r}, actual_time={!r})".format(
            self.__class__.__name__, self.cmd, self.timeout, self.actual_time
        )

    def __str__(self):
        return self.__repr__()


class CommandTimeoutError(FlakyTimeoutError):
    def __repr__(self):
        return (
            "{}: "
            "command {!r} took longer than expected "
            "(actual time {!r}s, expected time {!r}s)"
        ).format(self.__class__.__name__, self.cmd, self.actual_time, self.timeout)


class PromptTimeoutError(FlakyTimeoutError):
    def __repr__(self):
        return (
            "{}: command {!r} didn't finish after {!r}s (expected time {!r}s)".format(
                self.__class__.__name__, self.cmd, self.actual_time, self.timeout
            )
        )


def run(  # pylint: disable=keyword-arg-before-vararg
    command, quiet=False, *args, **kwargs
):
    output, status = pexpect.run(command, withexitstatus=1, *args, **kwargs)
    if status != 0:
        if not quiet:
            print(output)
        raise subprocess.CalledProcessError(status, command)
    return output



"""pytest-netdut is a pytest plugin to provide fixtures for the control networked DUTs.

The purpose is to running automated tests of software on networking hardware,
including FPGA applications like those built using Arista's FDK.

Assuming the device is accessible via SSH, run tests like this:

```bash
pytest --device=1.2.3.4
```

# The `dut` fixture

The `dut` fixture is created dynamically (by [`create`][pytest_netdut.create]) when the pytest-netdut plugin
is loaded, based on information passed to pytest on the CLI. By default, it refers to a number of sub-fixtures,
which are also dynamically created on first use.

These sub-fixtures of `dut` are:

1. ssh -- an instance of the [CLI][pytest_netdut.CLI] class for this DUT.
2. eapi -- an instance of the [EAPI][pytest_netdut.EAPI] class for this DUT.
3. xapi -- an instance of the [xapi][pytest_netdut.xapi] class for this DUT.
4. hostname -- the hostname of this DUT.
5. sku -- the SKU, as a string, determined by querying the DUT.

# Usage

A test which uses netdut might look like:

```python
def test_showver(dut):
    info = dut.eapi.sendcmd("show version")
    logging.info(f"DUT model was: {info['modelName']}")
```

this uses the dut.eapi fixture to run `show version` on the DUT and then logs the value of `modelName` from the result.

# Skipping

`pytest-netdut` implements a skipping mechansim based on the dut, by providing marks which annotate the desired
behaviour for the tests. Current marks are:

1. `eos` -- marks a test to run on EOS only.
2. `mos` -- marks a test to run on MOS only.
3. `skip_device_type(pattern)` -- marks a test to be skipped when the DUT's SKU matches the pattern.
4. `only_device_type(pattern)` -- marks a test to only be run when the DUT's SKU matches the pattern.

An example of a test which would be skipped on a particular SKU:

```python
@pytest.mark.only_device_type("^DCS-7130")
def test_showver(dut):
    assert dut.eapi.sendcmd("show version")['modelName'].startswith("DCS-7130")
```

# Utilities

`pytest-netdut` provides some other functionality to make it easier to write tests.

* The [wait_for][pytest_netdut.wait_for] fixture repeatedly calls a function until
    a condition becomes true, or a timeout expires. This is useful for waiting for
    the configuration to take effect once the commands have been sent.

# Pytest Command Line Options

`pytest-netdut` adds a command line option:

* `--device` -- specifies the hostname for the device. Netdut expects to be able to connect via ssh
    to this hostname.

"""
from typing import Callable
import pytest
from . import factories
from .utils import wait_for
from .wrappers import CLI, EAPI, eapi_enabled_fixture, xapi

__all__ = ["CLI", "EAPI", "create", "eapi_enabled_fixture", "wait_for", "xapi"]


def pytest_configure(config):
    """pytest_configure hook that adds markers used by netdut.

    Args:
        config (pytest.Config): The pytest config object.
    """

    # register markers
    config.addinivalue_line("markers", "eos: mark a test as eos only.")
    config.addinivalue_line("markers", "mos: mark a test as mos only.")
    config.addinivalue_line(
        "markers", "skip_device_type: mark a test as excluding a particular SKU regex."
    )
    config.addinivalue_line(
        "markers",
        "only_device_type: mark a test as only running on a particular SKU regex.",
    )

    # FIXME: not really part of netdut
    config.addinivalue_line("markers", "slow: slower than usual tests.")


def pytest_addoption(parser):
    """pytest_addoption hook adds options used by netdut.

    Netdut uses the following CLI/ini/config file options:
    * --device : configure the hostname for the switch, which can be used for SSH
    * --console: the address of a TCP socket which serves the device's serial console

    Args:
        parser (pytest.Parser): The pytest parser to add configuration options to.
    """
    group = parser.getgroup("netdut")
    group.addoption("--device", dest="dut_hostname", help="DUT hostname")
    group.addoption("--console", dest="dut_console_url", help="DUT console URL")


def create(name) -> Callable:
    """Dynamically create all fixtures required to implement "dut"

    Fixtures can be created for any named dut. The DUT fixture named `dut`
    is created by default when netdut is loaded. These DUT fixtures
    refer to other fixtures which may be used dynamically (e.g. no resources
    are used until required).

    This function creates the DUT fixture itself, as well as the other associated
    fixtures, like the EAPI fixture (used for methods like `dut.eapi.sendcmds()`).

    We assume that there is a pre-existing fixture called `{name}_info` (where
    {name} is the value of the parameter passed to create.). It should be a
    dictionary containing a host name.

    Args:
        name (str): The name of the top level DUT fixture - e.g. `dut`

    Returns:
        The DUT fixture which was created.
    """
    fixtures = [
        factories.create_dut_fixture(name),
        factories.create_eapi_fixture(name),
        factories.create_hostname_fixture(name),
        factories.create_skipper_fixture(name),
        factories.create_sku_fixture(name),
        factories.create_softened_fixture(name),
        factories.create_ssh_fixture(name),
        factories.create_xapi_fixture(name),
    ]

    for fixture in fixtures:
        globals()[fixture.__name__] = fixture

    return fixtures[0]


@pytest.fixture(scope="session")
def dut_info(pytestconfig) -> dict:
    """A fixture providing the DUT configuration passed via config to pytest-netdut.

    Args:
        pytestconfig (pytest.Config): The pytestconfig object representing the pytest configuration.

    Returns:
        A dictionary containing the hostname and console URL config parameters.
    """
    info = {
        "hostname": pytestconfig.getoption("dut_hostname"),
        "console_url": pytestconfig.getoption("dut_console_url"),
    }
    assert info["console_url"] or info["hostname"], "You must specify a device"
    return info


create("dut")



import pytest
import logging


@pytest.mark.eos
def test_that_runs_on_EOS(dut):
    "A test decorated so that it only runs on EOS"
    #    logging.info("Must be EOS!")
    assert True


@pytest.mark.mos
def test_that_runs_on_MOS(dut):
    "A test decorated so that it only runs on MOS"
    #    logging.info("Must be MOS!")
    assert True


@pytest.mark.only_device_type(
    "DCS-7130",
    reason="Demonstrating a 7130-only test",
)
def test_that_runs_on_7130(dut):
    "A test decorated so that it only runs on 7130 platforms"
    #    logging.info("Must be a 7130!")
    assert True


@pytest.mark.skip_device_type(
    "DCS-7130",
    reason="Demonstrating a non-7130 test",
)
def test_that_skips_7130(dut):
    "A test decorated so that it only runs on non-7130 platforms"
    #    logging.info("Must not be 7130!")
    assert True



import logging
import pytest

# A simplified demo
def test_showver(dut):
    info = dut.eapi.sendcmd("show version")
    logging.info(f"DUT model was: {info['modelName']}")


# A demo of an assertion, which we expect to fail
@pytest.mark.xfail()
def test_check_version(dut):
    info = dut.eapi.sendcmd("show version")
    assert info["version"] == "1.0.0"



def test_ssh_showver(dut):
    """Demonstrate using the `dut` fixture to determine the version via ssh.
    SSH returns a text output."""

    dut.ssh.sendcmds("show version")


def test_eapi_showver(dut):
    """Demonstrate using the `dut` fixture to check the OS version via eapi."""
    version_info = dut.eapi.sendcmd("show version")
    assert version_info["version"] != "1.0.0"


def test_x_showver(dut):
    """Demonstrate the X api using the `dut` fixture"""
    assert dut.x.show_version()["version"] != "1.0.0"



# -------------------------------------------------------------------------------
# - Copyright (c) 2021 Arista Networks, Inc. All rights reserved.
# -------------------------------------------------------------------------------
# - Author:
# -   fdk-support@arista.com
# -
# - Description:
# -   Convenience wrappers around CLI/EAPI.
# -
# -   Licensed under BSD 3-clause license:
# -     https://opensource.org/licenses/BSD-3-Clause
# -
# - Tags:
# -   license-bsd-3-clause
# -
# -------------------------------------------------------------------------------
import contextlib
import re
import logging
import pprint

from typing import Callable

import pyeapi
import pytest
import six

from . import px


def _splitcmds(cmds):
    if isinstance(cmds, six.string_types):
        return [cmd.strip() for cmd in cmds.strip().splitlines()]
    return cmds


def _eos_to_mos_translator(cmds):
    translations = [
        (r"interface ap1/(.*)", r"interface ap\1"),
        (r"l1 source interface ap1/(.*)", r"source ap\1"),
        (r"l1 source interface ap(.*)", "CAN NOT TRANSLATE"),
        (r"l1 source interface (.*)", r"source \1"),
        (
            r"l1 source mac",
            r"source mac",
        ),
        (r"no l1 source", r"no source"),
        (r"bash sudo cortina", "CAN NOT TRANSLATE"),
        (r"traffic-loopback source network device phy", r"loopback internal"),
        (r"traffic-loopback source system device phy", r"loopback"),
        (r"no traffic-loopback", r"no loopback"),
    ]

    commands = []
    for command in _splitcmds(cmds):
        for (before, after) in translations:
            matcher = re.match(before, command)
            if matcher:
                commands.append(matcher.expand(after))
                break
        else:
            # If we don't get a match, just append the original.
            commands.append(command)
    if commands != cmds:
        logging.debug("Before: %s, After: %s", repr(cmds), repr(commands))
    return commands


class CLI(px.CLI):
    """Extends pexpect.spawn (via px.CLI) to add support for interacting with
    an Arista style CLI."""

    def sendcmds(self, cmds):
        """Returns a list of results of running multiple commands on the CLI.

        Args:
            cmds (str | Iterable[str]): Either an iterable of commands to be run or a newline separated string containing one or more commands.
        """
        return [self.sendcmd(cmd) for cmd in _splitcmds(cmds)]


class EAPI:
    """A wrapper around pyeapi connections with px-like methods.

    Args:
        hostname (str): The hostname or IP address of the device to connect to.
        transport (str): The transport to use to connect.
    """

    def __init__(self, hostname, transport):
        self._conn = pyeapi.connect(host=hostname, transport=transport)
        self.translator = None

    def set_translator(self, translator):
        """Sets the function to be called on each command before it is passed to the CLI or CAPI/eAPI.

        This can be used, for example, to hide syntax differences between devices with different CLIs.

        Args:
            translator (Callable): The function to be called before on each command.
        """
        self.translator = translator

    def sendcmd(self, cmd, timeout=None):
        """Returns the deserialized JSON result of running a command via the CAPI/eAPI.

        Args:
            cmd (str): The command to be run.
        """
        if timeout:
            logging.debug("Timeout parameter is ignored by EAPI -- timeout=%f", timeout)
        if self.translator:
            cmds = self.translator(cmd)
        else:
            cmds = [cmd]
        return self._conn.execute(cmds)["result"][0]

    def sendcmds(self, cmds, timeout=None):
        """Returns a list of results of running multiple commands via the CAPI/eAPI.

        Args:
            cmds (str | Iterable[str]): Either an iterable of commands to be run or a newline separated string containing one or more commands.
        """
        if timeout:
            logging.debug("Timeout parameter is ignored by EAPI -- timeout=%f", timeout)
        if self.translator:
            cmds = self.translator(cmds)
        logging.info(pprint.pformat(cmds))
        return self._conn.execute(_splitcmds(cmds))["result"]


@pytest.fixture(scope="session", name="eapi_enabled")
def eapi_enabled_fixture(wait_for):
    """A session-scoped fixture that automatically enables the EAPI on the dut and returns a px-like wrapper around it."""

    @contextlib.contextmanager
    def _eapi_enabled(hostname, ssh, transport):
        if ssh.cli_flavor == "mos":
            ssh.sendcmds(
                """
                configure
                management http
                    no protocol secure
                management api
                    no shutdown
                end
            """
            )
        else:
            ssh.sendcmds(
                """
                configure
                management api http-commands
                    no shutdown
                    validate-output
                management http-server
                    protocol http
                end
            """
            )
            # Wait for services to init
            ssh.sendcmd("wait-for-warmup Capi CapiApp")

        def connect():
            # Disable pyeapi error logging while we try to connect.
            logger = logging.getLogger("pyeapi")
            old_level = logger.getEffectiveLevel()
            logger.setLevel(logging.CRITICAL)
            try:
                eapi = EAPI(hostname=hostname, transport=transport)
                eapi.sendcmd("show version")
                if ssh.cli_flavor == "mos":
                    eapi.set_translator(_eos_to_mos_translator)
                return eapi
            except (pyeapi.eapilib.ConnectionError, ConnectionRefusedError):
                return None
            finally:
                logger.setLevel(old_level)

        yield wait_for(connect)

        # FIXME: we should we restore the config rather than just disabling EAPI.
        if ssh.cli_flavor == "mos":
            ssh.sendcmds(
                """
                configure
                management api
                    shutdown
                management http
                    default protocol
                end
            """
            )
        else:
            ssh.sendcmds(
                """
                configure
                management http-server
                    no protocol http
                management api http-commands
                    shutdown
                    no validate-output
                end
            """
            )

    yield _eapi_enabled


def xapi(eapi):
    """Provides a pythonic wrapper around the cAPI/eAPI.

    Examples:
        >>> xapi.show_version()["version"]
        '4.26.1FX-7130'
    """

    class X:
        def __init__(self, cmds=None):
            self.cmds = cmds or []

        def __getattr__(self, name):
            return X([*self.cmds, name.replace("_", " ")])

        def __call__(self):
            return eapi.sendcmds(self.cmds)[-1]

        def __getitem__(self, key):
            return X([*self.cmds[:-1], f"{self.cmds[-1]} {key}"])

        def add(self, elem):
            return X([*self.cmds, elem])()

    return X()
